/*******************************************************************
*
*       Simulated Maximum Likelihood estimation of the
*       stochastic volatiliy model
*       version 1.2
*       February, 1996
*
*       Jon Danielsson (c)
*       Department of Economics
*       University of Iceland
*       101 Reykjavik
*       Iceland
*       e-mail: jond@hag.hi.is
*       ftp  hag.hi.is  
*       WWW  http://www.hag.hi.is/~jond
*       
*       please cite in published work and inform the author.
*
******************************************************************/

#include "svm.h"

int agis;
int OKL(double x);

#define PI 3.14159265358979323846

#define ISNAN isnan

#ifdef __hp9000s700 
#define ISNAN isnan
#endif

#ifdef _WIN32
#define ISNAN _isnan
#endif

int sml1::allocate_mem(void)
{
	bb1   	= new double[T+2];
	bb2   	= new double[T+2];
	bb3   	= new double[T+2];

	muy 	= new double[T+1];   // storage for MC Lik
	r   	= new double[T+2];
	ap   	= new double[T+2];
	q   	= new double[T+2];

	av   	= new double[T+2];
	x2   	= new double[T+2];


	psi   	= new double[N+1];   // storage for MC Lik
	eps   	= new double[N+2];
	lagy   	= new double[N+2];
	lagy1  	= new double[N+2];

	a = new double [T+1];
	b = new double [T+1];
	c = new double [T+1];
	as = new double [T+1];
	bs = new double [T+1];
	cs = new double [T+1];

	return (int) NULL;
} 
int sml1::deallocate_mem(void)
{
	delete [] bb1;
	delete [] bb2;
	delete [] bb3;

	delete [] muy;
	delete [] r;

	delete [] ap;
	delete [] av;
	delete [] q;

	delete [] x2;
	delete [] psi;
	delete [] eps;
	delete [] lagy;
	delete [] lagy1;

	delete [] cs;
	delete [] bs;
	delete [] as;
	delete [] c;
	delete [] b;
	delete [] a;

	return (int) NULL;
}
double Hmt;

//	This funciton loops through the AGIS iterations and calls
//	a separate function to estimate the likelihood and regression
//	at each step

double sml1::func(void)
{
	xr.init(T,N,data,omega,delta,nu,bb1,bb2,bb3);  // initialize regression
	xr.RegType=Reg;
	int i,j,res,t;    
	fstream out;
	error_code=OK;

       	for(t=1;t<=T;t++) x2[t] = SQR(data[t]);		// precalculate squared values
	for(t=1;t<=T;t++) muy[t] = omega;
	i 	= get_isd();			// get initial AGIS parameters

	if(i != OK) return i;

	for(j=1;j<=I;j++)	// AGIS loop
	{
		agis = j;
		res 	= lik1_func();		// run one iteration
//		cout << "  " << mpsi << endl;
		if(diffindex<STOP_AGIS) j = I+1;
		else {
			if(res != OK) { error_code=res;return 0.0;}
			res 	= get_isd();	// get AGIS parameters
			if(res != OK) { error_code=res; return 0.0; }
		}
	}	// end AGIS loop
	if(ISNAN(mpsi)!=0){ error_code=SERR5; return 0.0; }
	if(mpsi>1e100 || mpsi < -1e300){error_code=SERR4; return 0.0;}
//		cout << "  " << mpsi << endl;
	return mpsi;
}


//	This runs one step of the likelihiood

int sml1::lik1_func(void)
{
	int t,n,res;
	double sums, nusum, init,ly,d1,d2,yyy,ls;
	res	= OK;
	muy[0] = 0.0;
	for(t=1;t<=T;t++) muy[t] = omega;

// the variable sums is the constant factor

	init 	=  (double) T * (-0.5*llog(2.0 * PI)) ;
	sums = 0.0;
	nusum = llog(nu/ssqrt(1.0 - delta*delta)) + T*llog(nu);	
	for(t=0;t<=T;t++) sums +=  llog(av[t]);
	sums = sums - nusum + init - 0.5*(a0);

	diffindex = 0.0;
        normal r;	// initializes my random number class
	int iseed = (int) Seed;
	r.init(iseed);	// sets seed for random number class

	for(n=1;n<=N;n++)            psi[n] = 0.0;

	r(eps,N+1);		// draws random number vector
	r(eps,N+1);		// draws random number vector
	for(n=1;n<=N;n++)        lagy1[n] = 0.0;
	double *ltmp = new double[N+1];

	for(t=1;t<=T;t++)	// loop through time
	{
		r(eps,N+1);
		for(n=1;n<=N;n++)   // loop through simulations
		{
			lagy[n] = ap[t]+q[t]*lagy1[n]+av[t]*eps[n];
			ly = lagy[n];
			yyy = x2[t] /exp(ly) + ly;
			ltmp[n] = yyy;
			lagy1[n] = lagy[n];
			psi[n]  += yyy - bb1[t] - bb2[t]*ly - bb3[t]*ly*ly;
			xr.add(n,ly,yyy);    // collect cor regression
		}	// end loop simulations
		xr.run(t,agis,diffindex,bb1,bb2,bb3);	// do regression
	}	// end loop time
	delete [] ltmp;

// be careful in getting likelihood, if I want to avoid overflows
	ls =  sums-0.5*mean(psi,N);
	d1 = 0.0;
	for(n=1;n<=N;n++) d1 += exp(sums-0.5*psi[n] - ls) ;
	if(OKL(d1)!=(int) NULL) {return SERR3;}
	mpsi = llog(d1) + ls - llog((double) N);

	for(n=1;n<=N;n++)   {   
	
	psi[n] = exp(sums-0.5*psi[n]);
				}

	d2 = mean(psi,N);
	varpsi = 0.0;
	for(n=1;n<=N;n++)       varpsi += SQR(d2-psi[n]);
	varpsi /= (double) (N-1);
	return res;
}
//  This does the AGIS backwards iterations, see appendix in Danielsson (1994, 1995) 
int sml1::get_isd(void)
{
	int t,er=OK;

	double mu0,v02;

	double nu2	=  nu*nu;
	double Delta 	=  delta;
	double Delta2	=  Delta*Delta;

	mu0 	= 0.0;
	v02	= nu2 / ( 1.0 - Delta2);


	for(t=1;t<=T;t++) {
		a[t]	=  bb1[t];
		b[t]	=  bb2[t];
		c[t]	=  bb3[t];
	}
// The backwards iteration
	as[T]	= bs[T] = cs[T] = 0.0;		
	for(t=T;t>0;t--) {
		av[t]	= nu2 /(1.0 + nu2*(c[t]+cs[t]));
		ap[t]	= -0.5 *(b[t]+bs[t]-2.0* muy[t]/nu2)*av[t];
		q[t]	= Delta * av[t]/nu2;

		as[t-1]	= a[t] + as[t] - ap[t]*ap[t] / av[t] + muy[t]*muy[t]/nu2;
		bs[t-1]	= -2.0 * ap[t] * q[t] / av[t] + 2.0*muy[t]*Delta/nu2;
		cs[t-1]	= Delta2/nu2 - q[t]*q[t] / av[t];
	}
// Error check

	er = OK;
	for(t=1;t<=T;t++) {
		if(ISNAN(av[t])!=0) {er=SERR2; break;} 
		if(av[t]<0.0)  	{er=SERR2; break;} 
		if(av[t]>1e300)	{er=SERR2; break;} 
		if(er != OK) return er;
		av[t] = ssqrt( av[t]); 
	}
	av[0]	= v02 / ( 1.0 + cs[0] * v02) ;
	isdmu0	= (mu0 / v02 - bs[0] / 2.0) * av[0];
	a0	= as[0] + mu0*mu0 / v02 - isdmu0*isdmu0 / av[0];

	if(ISNAN(av[0])!=0)  	{er=SERR2; } 
	if(av[0]<0.0)  	{er=SERR2; } 
	if(av[0]>1e300)	{er=SERR2;} 
	if(er != OK) return er;
	av[0] = ssqrt( av[0]); 

	return er;
}

//	I want to know if I cannot take logs, hence...

int OKL(double x){
	if(ISNAN(x)!=0)  	return 11;
	if(x<=0.0)  	 	return 11;
	return (int) NULL;
}
