/*******************************************************************
*
*       Simulated Maximum Likelihood estimation of the
*       stochastic volatiliy model
*       version 1.2
*       February, 1996
*
*       Jon Danielsson (c)
*       Department of Economics
*       University of Iceland
*       101 Reykjavik
*       Iceland
*       e-mail: jond@hag.hi.is
*       ftp  hag.hi.is  
*       WWW  http://www.hag.hi.is/~jond
*       
*       please cite in published work and inform the author.
*
******************************************************************/


#include <iostream.h>
#include <iomanip.h>
#include <math.h>
#include <fstream.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <float.h>
#include <string.h>
#include "timer.h"

//	library headers
#include "random.h"
#include "optimize.h"


//        DEFINES   
#define STOP_AGIS 1e-4
#define SQR(a) ((a)*(a))

#define SVM_VERSION "1.2"
#define SVM_DATE "Febrary 15, 1996"


#define err(x) error(__LINE__,__FILE__,x)
void error(int line, char* file, char* text);
double var(double *x, int num);
double mean(double *x, int num);

#define llog(x) eelog(x,__LINE__,__FILE__)
double eelog(double x ,int l,char *name);
#define ssqrt(x) eesqrt(x,__LINE__,__FILE__)
double eesqrt(double x ,int l,char *name);

enum {PRINT,NOPRINT};
enum {OK,SERR1,SERR2,SERR3,SERR4,SERR5};
enum {YES,NO};

//	This class does regression of Q
class regress {
private:
	double aa,bb,cc,dd,ee,x1y,x2y,x3y;	// used in manual regression
						// see below

	double* X;			// X vector in regression
	double* Y;			// y vector in regression

// 			These are used in singular value decomposition
	double** uu;
	double** vv;
	double *w,*b, wmax,wmin;	

//			coefficients from last iteration
	double* lb1,* lb2,* lb3;

//			coeff. values from regression
	double* bd1,* bd2,* bd3;
	double *bd;

	int T;				// num. obs.	
	int N;				// num. simualtions
	int ii;				// index stuffing X vector
	int all;			// has memory been allocated?

public:
	regress(){all=0;}
	int RegType;
	init(int t, int n,double* data, 
		double omega,double delta, double nu,
		double *bb1,double *bb2,double* bb3);
	~regress(void);
	void add(int n, double llik,double yyy);  
	void run(int t,int iter,double& diffindex,
		double *bb1,double *bb2,double* bb3);
};


// This parses the command line and reads the input file
class command_line {
protected:
	void read_file(void);
	void head(void);
	void foot(void);
public:
	char Fname[100];		// Input file
	int N;
	int T;
	int I;				// AGIS iterations
	int H;				// Calculate hessian Yes/No
	double P;			// Precision
	double Z;			// Hessian presicion
	double Seed;			//
	int Print;			// Print control 
	double omega;			//
	double delta;			// 
	double nu;			// 
	int Reg;			//  robust regression, YES

	double Hess_omega;		//
	double Hess_delta;		// 
	double Hess_nu;			// 
	double maxvalue;		// value at top;
	int Ascale;			// scale the mean
	double scale;

	parse(int argc,char* argv[]);
	double *data;
	double *raw_data;
};

//	This is the workhorse class

class sml1 : public command_line {
private:
	int final_opt;
	double diffindex;
	int 	optimize();
	int     deallocate_mem(void);


        double varpsi;
	double mpsi;		// validated 
	double isdmu0;		// The initial mean for ISD density 
	double init;		// The constant part of Psi 
	double a0;
// ***********************************************************

	int lik1_func(void);
	int get_isd(void);
// ********************************* 

	double* q;      // The Est q for calculation of ISD means
	double* ap;     // The Est p for calculation
//	double* axv;    // The Est of variance of ISD
	double* av;     // The Est of variance of ISD
	double* r;      // The Estimator of variance of ISD
	double* x2;     // Values of x*x
	double* psi;
	double* eps;
	double* lagy;
	double* lagy1;
	double* bb1,* bb2,* bb3;	// regression coefficients (Q)

//	double* lb1, double* lb2; double* lb3;
	double* muy;            // Deterministic Mean vector of y process
	double *a, *b, *c,*as,*bs,*cs;	// used in get_isd()

public:
	regress xr;
        int     allocate_mem(void);
	void footer(void);
	sml1(){};
	~sml1(){
		deallocate_mem();
	}
	int Run(void) {
		optimize();
		return (int) NULL;
	}
	double	OptFunc(double *x);

	int error_code;
	double func(void);
};


